#!/usr/bin/perl
#
# acmemail - An IMAP/POP3 to Web gateway
#
# See http://www.astray.com/acmemail/
#
# acmemail is Copyright (c) 1997-2000 Leon Brocard. All rights reserved.
# portions Copyright (c) 2001 Peter Watkins
#
# You may distribute acmemail under the terms of either the GNU General
# Public License as published by the Free Software Foundation (either
# version 2 of the License or at your option any later version) 
# or the Artistic License. You should have received a copy of both
# licenses along with this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# Version: see acme_version() in AcmemailConf.pm

# Include any library directories

use lib '.';
#use lib '/lib/place';

# There's no reason for other users to read any files
# that are written by acmemail (thanks to Deepak Gupta)
umask(077);

# Let's import a whole load of modules, including AcmemailConf
# which is the user-changable one

use strict; # Disabled for the released version!
use lib '.'; # For taint
use OneTimePadPW qw(GetRandomBytes EncryptOTP DecryptOTP);
use Digest::MD5;
use AcmemailConf qw(image_path mailhost smtphost hide_email_address mailhost_type local_domain
		    max_messages_per_page show_recent_messages_first
		    temporary_mime_directory web_temporary_mime_directory
		    session_type dope_directory session_timeout
		    inbox_directory force_cookies force_nojavascript 
		    force_defang_html force_escape_html safe_html_list
		    auto_add_hyperlinks referer_header_test
		    html_login acme_server_name acme_server_base
		    secure_cookie cookie_path control_cookie_path
		    acme_anonymize_links control_port message_port
		    acme_mem_parse_buffer acme_preferred_msg_type
		    html_head html_bottom html_bar html_message_bar
		    html_list_head html_list_row html_list_bottom
		    html_message_head html_message_headers html_message_bottom
		    html_quoted html_signature encrypt_password);
use AcmemailSession;
use CGI 2.47 qw(:standard);
use File::Basename;
use HTML::Entities qw(encode_entities);
use Mail::Internet;
use Mail::Cclient 1.5 qw(set_callback);
use MIME::Entity;
use MIME::Parser;
use Net::SMTP;
use Date::Format;
use Date::Parse;
use URI::Escape;
use URI::WithBase;
use CGI::Carp qw(fatalsToBrowser);
use vars qw($USE_MOD_PERL);


# ------------------- Stuff for different versions of perl --------------------

# Perl's exit() built-in function cannot be used in mod_perl scripts.
# Calling it causes the server child to exit (which makes the whole 
# idea of using mod_perl irrelevant). Here we use the Apache::exit()
# function instead, if we're under mod_perl

BEGIN {
  # Auto-detect if we are running under mod_perl or CGI.
  $USE_MOD_PERL = ( (exists $ENV{'GATEWAY_INTERFACE'}
                     and $ENV{'GATEWAY_INTERFACE'} =~ /CGI-Perl/)
		    or exists $ENV{'MOD_PERL'} ) ? 1 : 0;
}

#####
# Quit gracefully (if we're under mod_perl), otherwise exit()
#####

sub bye {
  my $output = shift;
  if (defined $output) {
    warn "acmemail error: $output";
    print qq|<p><font color="red">acmemail error: $output</font><p>|;
  }
  $USE_MOD_PERL ? Apache::exit(0) : CORE::exit(0);
}


# ------------------------------ Parameter extraction ---------------------------

# Who needs STDOUT caching?
$| = 1;

# A new CGI object
my $q = new CGI;

# Get all the CGI args we want
# (twenty-four parameters atm)
my $username = $q->param('username');
my $password = $q->param('password');
my $origmailbox = $q->param('mailbox');
my $offset = $q->param('offset');
my $show = $q->param('show');
my $delete = $q->param('delete.x');
my @todelete = $q->param('todelete');
my $logout = $q->param('logout');
my $msgkey = $q->param('msgkey');
my $compose = $q->param('compose');
my $send = $q->param('send');
my $to = $q->param('to');
my $cc = $q->param('cc');
my $bcc = $q->param('bcc');
my $subject = $q->param('subject');
my $body = $q->param('body');
my $attachment = $q->param('attachment');
my $reply = $q->param('reply');
my $replytoall = $q->param('replytoall');
my $forward = $q->param('forward');
my $inreplyto = $q->param('inreplyto');
my $aredir = $q->param('aredir');
my $mb = $q->param('mb');
my $att = $q->param('att');

# is this a dangerous request?
my $isDangerous = 0;
if ( 
	#  delete - deleting messages
	defined($delete) || 
	#  logout - a real hassle if composing a message,
	#	    and an annoyance in other situations
	defined($logout) || 
	#  sent $username and $password means they're trying to log in
	(defined $username and defined $password) ||
	#  send   - don't want CSRF attacks spoofing messages
	defined($send) ) { 
		# This action may merit special precautions
		$isDangerous = 1; 
}

# are we being asked to display unsafe content?
my $untrustedContent = 0;
if (defined($msgkey)) { $untrustedContent = 1; }

# what our referer should look like
my $refer;
my $safeBase = my $unsafeBase = lc(acme_server_base());
if ( control_port() ne '' ) {
  $safeBase =~ s/\:\d*$//;
  if ( ( ($ENV{HTTPS} eq "on") && (control_port() ne '443') ) ||
       ( ($ENV{HTTPS} ne "on") && (control_port() ne '80') ) ) {
    $safeBase .= ':'.control_port();
  }
}
if ( message_port() ne '' ) {
  $unsafeBase =~ s/\:\d*$//;
  if ( ( ($ENV{HTTPS} eq "on") && (message_port() ne '443') ) ||
       ( ($ENV{HTTPS} ne "on") && (message_port() ne '80') ) ) {
    $unsafeBase .= ':'.message_port();
  }
}
if ( $isDangerous == 1 ) {
  $refer = $safeBase. $ENV{'SCRIPT_NAME'} . '/control/';
} else {
  $refer = $unsafeBase. $ENV{'SCRIPT_NAME'};
}

# what port they connected to us on
my $requestPort = $ENV{'HTTP_HOST'};
$requestPort =~ s/.*\:(\d{1,})/$1/;
if (! ($requestPort =~ /^\d{1,}$/) ) {
  $requestPort = ($ENV{HTTPS} eq "on") ? '443' : '80';
}

# for the message display
my %msgtops;
my %msgtypes;
my $msgbottom;
my $noInline = 1;
my %inlineImages;
my %imageName;
my %imageSize;
my %imageURI;
my %shownImages;
my $attachNum = 0;

my $mailbox = $origmailbox;
my $inbox = inbox_directory();
if (not defined $mailbox) {
  $mailbox = "INBOX";
} elsif ($mailbox eq 'INBOX') {
  1; # do nothing
} elsif ($mailbox !~ /^$inbox/) {
  $mailbox = "$inbox/$mailbox";
}


# Here is the default paranoid HTTP header
my %http_header = (-expires => 'now',
		   'Cache-Control' => 'no-cache, must-revalidate',
		   -pragma => 'no-cache');

# By default we try and see if the browser will accept a cookie
# (but don't bother testing if they're already giving us cookies!)
if (! defined $q->cookie('session-id') ) {
	$http_header{-cookie} = cookie(-name=>'cantakecookies',
			       -value=>1,
			       -path=>cookie_path(),
			       -secure=>secure_cookie()); 
}

# Get the session id out of the cookie if it's there, or otherwise
# from a CGI param
my $sessionid = $q->cookie('session-id') if defined $q->cookie('session-id');

my $passwordPad = '';
$passwordPad = $q->cookie('key') if defined $q->cookie('key');

my $controlHash = '';
my $controlValue = '';
$controlValue = $q->cookie('control') if defined $q->cookie('control');

my $messageSecret = '';

if (defined $q->param('sessionid')) {

  # Wait, I told you to force cookies!
  if (force_cookies()) {
    print header(\%http_header);
    print html_head($q);
    print "<b>Your must accept cookies to use acmemail - you recently used a CGI parameter. ";
    print "Please re-login once you have turned them on.</b><br>";
    print html_login($q,$safeBase,$unsafeBase);
    print html_bottom($q);
    bye;
  }

  # Get the sessionid then
  $sessionid = $q->param('sessionid') if defined $q->param('sessionid');
  $passwordPad = $q->param('key') if defined $q->param('key');
  $controlValue = $q->param('control') if defined $q->param('control');
}


# Let's worry about the attachment now...
my $attachment_type;
{
  my $fh = $q->upload('attachment');
  if (defined $fh) {
    $attachment .= $_ while <$fh>;
    my $filename = $q->param('attachment');
    $attachment_type = $q->uploadInfo($filename)->{'Content-Type'};
  }
}


# ---------------- Security Checks ------------------------------

my $session; # Persistant object for security

AcmemailSession->set_type(session_type());

if ( ($sessionid =~ /[^a-z0-9]/) ) {
  # at various spots we assume $sessionid is a safe value for
  # things like file/directory names
  print header(\%http_header);
  $q->delete_all();
  print html_head($q);
  print html_bar($q);
  print "<h3>Invalid session id</h3>\n";
  print "Please re-login<br>";
  print html_login($q,$safeBase,$unsafeBase);
  print html_bottom($q);
  bye;
} elsif (defined($aredir) && $aredir =~ /^(http|https):\/\//i) {
  # make an anonymous redirect
  print header(\%http_header);
  print '<html><head><title>acmemail: anonymous redirect</title>';
  print '<meta http-equiv=refresh ';
  print 'content="1; url=' . CGI::escapeHTML($aredir) . "\">\n";
  print '</html><body>Attempting to redirect to ';
  print CGI::escapeHTML($aredir).'</body></html>';
  bye;
} elsif (
    ($untrustedContent && (message_port() ne '') && ($requestPort != message_port()) ) ||
    ((! $untrustedContent ) && (control_port() ne '') && ($requestPort != control_port()))
  ) {
  # need to change ports
  if ( $ENV{'REQUEST_METHOD'} eq 'GET' ) {
    # we can simply redirect the user
    my $correctURL = $q->url(-full=>1, -query=>1, -path_info=>1);
    my $correctPort = ':'. ($untrustedContent ? message_port() : control_port());
    if ( ( ($ENV{HTTPS} eq "on") && ($correctPort eq '443') ) ||
         ( ($ENV{HTTPS} ne "on") && ($correctPort eq '80') ) ) {
      $correctPort = '';
    }
    $correctURL =~ s/^([^\/]*\/\/[^\/]*?)(:\d{1,}|)\//${1}::PORTHERE::\//;
    $correctURL =~ s/::PORTHERE::/$correctPort/e;
    print $q->redirect($correctURL);
    bye;
  }
  print header(\%http_header);
  $q->delete_all();
  print html_head($q);
  print html_bar($q);
  print "<h3>Invalid URL</h3>\n";
  print "You have used an invalid URL. URLs for messages and attachments ";
  print "should use port ".message_port()." while other pages should have ";
  print "port ".control_port()." in their URLs.<br>";
  print html_bottom($q);
  bye;
} elsif (defined($show) && ($q->path_info() ne '')) {
  # do not display messages with PATH_INFO values,
  # to prevent XSS theft of the control cookie
  print $q->redirect($q->url(-full=>1, -query=>1, -path_info=>0));
  bye;
} elsif ( defined($att) && ( (!defined($msgkey)) || (! ($q->path_info() =~ /^\/view\//)) ) ) {
  # requesting to view an attachment but bad path
  print header(\%http_header);
  $q->delete_all();
  print html_head($q);
  print html_bar($q);
  print "<h3>Invalid request</h3>\n";
  print "<p>You have requested an attachment with an invalid URL.</p>\n";
  print "<p>You might wish to use the <a href=\"${refer}?logout=1\">logout function</a> and then <a href=\"${refer}\">log in</a> again.</p>";
  print html_bottom($q);
  bye;
} elsif ( ((!defined($show)) && (!defined($aredir)) && (!defined($att))) && $q->path_info() ne '/control/') {a
  # pages other than the message view page should be called with a 
  # /control/ URL
  $q->path_info('/control/');
  print $q->redirect($q->url(-full=>1, -query=>1, -path_info=>1));
  bye;
} elsif (referer_problem($ENV{'HTTP_REFERER'},$isDangerous,referer_header_test(),$refer)) {
  # failed referer security test, notify, link to login
  print header(\%http_header);
  $q->delete_all();
  print html_head($q);
  print html_bar($q);

  print "<p>Your request did not present a valid HTTP \"Referer\" header. It therefore is not possible for acmemail to verify that the request you have made is legitimate, and not an attempt to foil acmemail security.";

  print "<p>acmemail expected to see \"$refer\" but your client presented the Referer \"".CGI::escapeHTML($ENV{'HTTP_REFERER'})."\".";

  print "<p>You might wish to use the <a href=\"${refer}?logout=1\">logout function</a> and then <a href=\"${refer}\">log in</a> again.";

  print html_bottom($q);
  bye;
} elsif (defined $username and defined $password) {

  if (force_cookies() and not defined $q->cookie('cantakecookies')) {
    print header(\%http_header);
    $q->delete_all();
    print html_head($q);
    print "<b>Your must accept cookies to use acmemail. This is for security considerations. ";
    print "Please re-login once you have turned them on.</b><br>";
    print html_login($q,$safeBase,$unsafeBase);
    print html_bottom($q);
    bye;
  }

  # Create a new session
  # figure out the value for the control cookie  
  $controlValue = OneTimePadPW::GetRandomBytes(16);
  $messageSecret = OneTimePadPW::GetRandomBytes(16);
  my $md5 = new Digest::MD5;
  $md5->add($controlValue);
  $controlHash = $md5->hexdigest();
  my $storepassword = $password;
  if ( encrypt_password() ) {
	# How long should the key be? Use N * 32 bytes so we have
	# enough to encrypt the whole password, regardless of length
	# We use multiples of 32 to hide the true length of the password,
	# though obviously a key of 64 bytes means a password > 32 chars, etc. ;-)
	# NOTE: we should probably error out if it's very long, e.g. >= 128 bytes,
	#       as very long cookies may not work properly
	my $padLength =  ( int( 1 + (length($password) / 32) ) * 32);
	$passwordPad = OneTimePadPW::GetRandomBytes($padLength);
	$storepassword = OneTimePadPW::EncryptOTP($password,$passwordPad);
  }
  $session = AcmemailSession->create_new($username, "${controlHash}\t${messageSecret}\t$storepassword");
  $sessionid = $session->id;

  # set domain so session/key cookies go to all ports 
  my $domain = $ENV{'HTTP_HOST'};
  $domain =~ s/\:.*$//;
  # not needed if using same port for everything
  if ( control_port eq '' && message_port() eq '') { $domain = undef; }

  $http_header{-cookie} = 
      [ cookie(-name=>'control', -value=>$controlValue,
        -secure=>secure_cookie(), -path=>control_cookie_path($q)),
      cookie(-name=>'session-id', -value=>$sessionid, -domain=>$domain,
        -secure=>secure_cookie(), -path=>cookie_path()) ];
  
  if ( encrypt_password() ) {
    $http_header{-cookie} = 
      [ cookie(-name=>'key', -value=>$passwordPad, -domain=>$domain,
        -secure=>secure_cookie(), -path=>cookie_path()),
      cookie(-name=>'control', -value=>$controlValue,
        -secure=>secure_cookie(), -path=>control_cookie_path($q)), 
      cookie(-name=>'session-id', -value=>$sessionid, -domain=>$domain,
        -secure=>secure_cookie(), -path=>cookie_path()) ];
  }
} else {
 if (defined $sessionid) { 

  # Recreate an existing session
  $session = AcmemailSession->new($sessionid);

  if (defined $session) {
    $username = $session->username;
    my $passbundle = $session->password;
    $passbundle =~ /^([^\t]*)\t([^\t]*)\t(.*)$/;
    $controlHash = $1; $messageSecret = $2; $password = $3;
    # If we have a key, we need to decrypt the stored password value
    if ( $passwordPad ne '' ) {
	$password = OneTimePadPW::DecryptOTP($password,$passwordPad);
    }
  } else {
    # Problem?
    # Print the header (with paranoid expires and possibly cookies)
    print header(\%http_header);

    # And our own little header
    print html_head($q);
    print "<b>Your connection may have timed out. Please re-login.</b><br>";
    print html_login($q,$safeBase,$unsafeBase);
    print html_bottom($q);
    bye;
  }

 } else {

  # No access, demand they identify themselves immediately
  # Print the header (with paranoid expires and possibly cookies)
  print header(\%http_header);

  # And our own little header
  print html_head($q);
  print "Please log in.<br>\n";
  print html_login($q,$safeBase,$unsafeBase);
  print html_bottom($q);
  bye;
 }

 if ( ($q->path_info() eq '/control/') && control_cookie_problem() ) {
  # failed control cookie security test, notify, link to login
  print header(\%http_header);
  $q->delete_all();
  print html_head($q);
  print html_bar($q);

  print "<p>Your request did not present valid control credentials. It therefore is not possible for acmemail to verify that the request you have made is legitimate, and not an attempt to foil acmemail security.";

  print "<p>You might need to log in again.";
  print html_login($q,$safeBase,$unsafeBase);

  print html_bottom($q);
  bye;
 } elsif ( defined($show) && msgkey_problem() ) {
  # failed control cookie security test, notify, link to login
  print header(\%http_header);
  print html_head($q);
  print html_bar($q);
  $q->delete('show');
  $q->delete('msgkey');
  $q->path_info('/control/');
  $q->param('list','1');

  print "<p>Your request did not present a valid message key. It therefore is not possible for acmemail to verify that the request you have made is legitimate, and not an attempt to foil acmemail security.\n";

  print "<p>To get a valid message key, start at the <a href=\"".$q->self_url()."\">message list page</a>.\n";

  print "<p>You might need to log in again.";

  print html_bottom($q);
  bye;
 } elsif (defined $logout) {

  # Revoke access by expiring the session
  $session = AcmemailSession->new($sessionid);
  $session->expire if defined $session; 
  # Now we need to expire the cookie (if there is one) too
  $http_header{-cookie} = [ cookie(-name=>'session-id',
				 -value=>'',
				 -secure=>secure_cookie(),
				 -path=>cookie_path(),
				 -expires=>'now'),
  # And the password key
			    cookie(-name=>'key',
				 -value=>'',
				 -secure=>secure_cookie(),
				 -path=>cookie_path(),
				 -expires=>'now'),
  # And the control key
			    cookie(-name=>'control',
				 -value=>'',
				 -secure=>secure_cookie(),
				 -path=>control_cookie_path($q),
				 -expires=>'now') ];

  print header(\%http_header);
  $q->delete_all();
  print html_head($q);

  print html_login($q,$safeBase,$unsafeBase);
  print html_bottom($q);
  bye;
 }
}

if ( defined($att) ) {
  # display attachment?
  # test checksum
  my $md5 = new Digest::MD5;
  $md5->add($att,$messageSecret);
  if ( $msgkey ne $md5->hexdigest() ) {
    # invalid checksum
    $q->delete_all();
    print header(\%http_header);
    print html_head($q);
    print "<h3>Invalid request</h3>\n";
    print "<p>You have requested an attachment with an invalid URL.</p>\n";
    print "<p>You might wish to use the <a href=\"${refer}?logout=1\">logout function</a> and then <a href=\"${refer}\">log in</a> again.</p>";
    print html_bottom($q);
    bye;
  }
  my $fpath = dope_directory().$sessionid.'-mime/'.acme_file_escape($att);
  my @finfo = stat($fpath);
  if ( ($? != 0) || (! open(F,"<$fpath")) ) {
    # cannot find file!
    $q->delete_all();
    print header(\%http_header);
    print html_head($q);
    print "<h3>No such attachment!</h3>\n";
    print "<p>acmemail cannot find the attachment you requested. You may need to go back and reload the message display page to make the attachment available.</p>\n";
    print html_bottom($q);
    bye;
  }
  my $fsize = $finfo[7];
  my $mheader = '';
  my $fbuff = '';
  my $gotmhead = 0;
  while ( ($gotmhead == 0) && (read(F,$fbuff,1)) ) {
    if ( $fbuff ne chr(0) ) {
      $mheader .= $fbuff;
    } else { $gotmhead = 1; }
  }
  $fsize -= (length($mheader) + 1);
  %http_header = (-expires => 'now',
		   'Cache-Control' => 'no-cache, must-revalidate',
		   -pragma => 'no-cache',
		   -content_type => $mheader,
		   -content_length => $fsize);
  print header(\%http_header);
  while ( read(F,$fbuff,2048) ) { print $fbuff; }
  close F;
  bye;
}

# Print the header (with paranoid expires and possibly cookies)
print header(\%http_header);

if ( defined($show) && (!defined($q->cookie('session-id'))) && (!defined($mb)) ) {
  # They want a message page but present a control argument;
  # give them a frameset with a safe message area
  $q->param('mb','0');
  my $topurl = CGI::escapeHTML($q->url(-full=>1, -query=>1, -path_info=>1));
  $q->param('mb','1');
  $q->delete('control');
  my $bottomurl = CGI::escapeHTML($q->url(-full=>1, -query=>1, -path_info=>0));
  print <<EOPAGE;
<html><head><title>acmemail</title></head>
<frameset rows="165,*" border="0" bordercolor="#ffffff">
 <noframes>
 For security reasons, acmemail uses frames if your browser does
 not accept cookies. We strongly recommend accepting cookies 
 for acmemail.
 </noframes>
 <frame name="controls" src="${topurl}" scrolling="no" frameborder="1" bordercolor="#000000" marginheight="0" marginwidth="0">
 <frame name="message" src="${bottomurl}" frameborder="0" bordercolor-="#ffffff" marginheight="0" marginwidth="0">
</frameset>
</html>
EOPAGE
  bye;
}

# And our own little header
if ( (!defined($mb)) || ($mb eq '0')) {
  print html_head($q);
}

# And then delete all CGI params for good measure
$q->delete_all();


# Let the sessionid live on
if ( (!force_cookies()) && (!defined($q->cookie('cantakecookies'))) ) {
  if (!defined($q->cookie('session-id'))) {
    $q->param('sessionid', $sessionid);
  }
  if ( (!defined($q->cookie('key'))) && encrypt_password() ) {
    $q->param('key', $passwordPad);
  }
  if( (!defined($q->cookie('control'))) && ($q->path_info() eq '/control/') ) {
    $q->param('control', $controlValue); 
  }
}

# And possibly the mailbox too...
$q->param('mailbox', $origmailbox);

# Oh, and the offset (if there is one)...
$q->param('offset', $offset);

# -------------------------- Contact the mail server --------------------------

# Set IMAP login callback
set_callback(login => sub { return ($username, $password) });

# Set some global parameters: prevent rsh attempts and
# decrease the maximum number of login trials to 1
Mail::Cclient::parameters(undef, RSHTIMEOUT => 0, MAXLOGINTRIALS => 1);

# Connect to the mail server using whatever protocol is defined in config file
my $cclient;
$cclient = Mail::Cclient->new('{' . mailhost() . '/' . mailhost_type() . '}' . $mailbox);

# If we couldn't connect, complain in a loud voice then stop dead
unless (defined $cclient) {
  print "<b>Cannot login with current username/password - please try again.</b><br>";
  print html_login($q,$safeBase,$unsafeBase);
  print html_bottom($q);
  bye;
}


# -------------------- Print the rest of the web page ------------------------
 
# print the 'toolbar'
if (defined $show) {
  if ($mb eq '1') {
    my $image_path = image_path();
    print <<EOHEAD;
<meta http-equiv="Content-Type" content="text/html; charset=ISO-8859-1">
<title>acmemail</title>
<LINK rel="stylesheet" type="text/css" href="${image_path}acmemail.css">
<base target="_top">
</head>
<body bgcolor="#ffffff">
EOHEAD
  } elsif ( $mb eq '0' ) {
    $q->param('control',$controlValue);
    print html_message_bar($q, $show);
    bye;
  } else {
    print html_message_bar($q, $show);
  }
} else {
  print html_bar($q);
}

# switch mode and print the main body depending entirly on what we're up to
if    (defined $show)    { show($q, $cclient, $session, $show)                 }  
elsif (defined $delete)  { trash($q, $cclient, @todelete);
			   list($q, $cclient, $mailbox, $offset);              }
elsif (defined $compose) { compose($q, $cclient, $to, $cc, $bcc, $inreplyto, $subject,
				   $attachment, $attachment_type, $body)       }     
elsif (defined $send)    { mail($q, $cclient, $mailbox, $offset, $username,
				$to, $cc, $bcc, $inreplyto, $subject, $attachment, 
				$attachment_type, $body)                       }
elsif (defined $forward) { forward($q, $cclient, $forward)                     }   
elsif (defined $reply)   { reply($q, $cclient, $reply, $replytoall)            }     
else                     { list($q, $cclient, $mailbox, $offset)               }

# Finish off with closing tags
print html_bottom($q);

# And finally close the mail connection
$cclient->close;

# to be sure, exit
bye;




# ----------------------------------------------------------------------
#
#   From here on, we have all the bits that perform most of the 
#   functionality of Acmemail.
#





#####
# Show the list of messages
#####

sub list($$$$) {
  my($q, $cclient, $mailbox, $offset) = @_;

  my $nmsgs = $cclient->nmsgs;

  my @messages = (1..$nmsgs);
  @messages = reverse @messages if show_recent_messages_first();

  my ($previous_offset, $next_offset);
 
  if (max_messages_per_page() != 0) {
    @messages = splice @messages, $offset, max_messages_per_page();
    $previous_offset = $offset - max_messages_per_page() if ($offset - max_messages_per_page() >= 0);
    $next_offset = $offset + max_messages_per_page() if ($offset + max_messages_per_page() < $nmsgs);
  }

  print html_list_head($q, $username, $nmsgs, $previous_offset, $next_offset, $mailbox, [mailboxes($cclient)]);

  # And for every message...
  foreach my $i (@messages) {
    my $uid = $cclient->uid($i);
    my $uid2 = $cclient->uid($i);
    $uid = $i; # Quick hack until we properly fix UID's. Oops.

    my $subject = $cclient->fetchheader($i, ['Subject']);
    $subject =~ s|^Subject: +||;
    $subject = quote_subject($q, $subject);

    my $from = $cclient->fetchheader($i, ['From']);
    $from =~ s|^From: +||;
    $from = quote_from($q, $from);

    my $date = $cclient->fetchheader($i, ['Date']);
    $date =~ s|^Date: +||;

    # date must be suitable long enough.
    # this test takes care of the case where there is
    # no Date header.
    $date = quote_date($q, $date) if length($date) > 10;

    my ($head, $body) = $cclient->fetchstructure($i);
    my $elt = $cclient->elt($i);

    # Mark notseen flag if the message has not been seen yet
    my $notseen = not grep {$_ eq '\Seen'} @{$elt->flags};

    my $size = int ($elt->rfc822_size / 1024);
    1 while $size =~ s/(.*\d)(\d\d\d)/$1,$2/;
    $size =~ s/^0$/\&lt;1/;
    $size .= "K";

    # Mark attachment flag true if there is one
    my $attachment;
    $attachment = 1 if ((lc $body->type eq 'multipart') and (lc $body->subtype eq 'mixed'));

    # And print out the row
    print html_list_row($q, $uid, $subject, $from, $date, $size, $notseen, $attachment, $uid2, $messageSecret, $unsafeBase);
  }

  if ( (!force_cookies()) && (!defined($q->cookie('cantakecookies'))) ) {
    if (!defined($q->cookie('session-id'))) {
     print $q->hidden('sessionid',$sessionid);
    }
    if ( (!defined($q->cookie('key'))) && encrypt_password() ) {
      print $q->hidden('key',$passwordPad);
    }
    if( !defined($q->cookie('control')) ) {
      print $q->hidden('control',$controlValue);
    }
  }

  # Finish off the table and the form
  print html_list_bottom($q);
}






######
# Show an individual message
######

# Just retrieve it
sub show($$$$) {
  my($q, $cclient, $session, $i) = @_;
  my $message = $cclient->fetchheader($i,'uid');
  $message .= $cclient->fetchtext($i,'uid');

  if ($message) {

    # Create and set the output directory:
    my $mimedir = temporary_mime_directory() . $session->id;

    (-d $mimedir ) || mkdir "$mimedir", 0700 || bye "mkdir: $!";
    (-w $mimedir ) || bye "can't write to temporary mime directory ($mimedir): $!";

    show_mime($mimedir, $message);

  } else {
    print "<h3>Error fetching message</h3>Maybe it doesn't exist any more?";
  }
}


# Parse it as a MIME message;
sub show_mime($$) {
  my($mimedir, $message) = @_;

  my $parser = new MIME::Parser;

  $parser->output_to_core(acme_mem_parse_buffer());
  my $tdir = dope_directory().$sessionid.'-mime';
  (-d $tdir ) || mkdir "$tdir", 0700 || bye "unable to make session temp dir";
  $parser->output_dir($tdir);
  $parser->parse_nested_messages('REPLACE');
  my $entity = $parser->parse_data($message) or bye "Parse error! Error while paring MIME message.";

  print html_message_head($q);
  
  # Dump it out
  dump_entity($entity, 'email');

  my $msgtop;
  my $preftype = '';
  foreach my $mtype ( sort keys %msgtypes  ) {
    if ( $mtype eq acme_preferred_msg_type() ) {
      $preftype = $mtype;
    }
    $msgtop .= $msgtops{$mtype}."<br><hr><br>\n";
  }
  if ( $preftype ne '' ) { 
    $msgtop = $msgtops{$preftype}; 
  }
    
  $msgtop = acme_fix_cid_src($msgtop);
  print $msgtop;
  foreach my $tag (keys %inlineImages) {
    if ( $shownImages{$tag} != 1 ) {
      # not displayed in=line, put in table
      my $bpath = $imageURI{$tag};
      $msgbottom .= "<br><table bgcolor=\"#cccccc\" cellspacing=0 cellpadding=2 border=0%>";
      $msgbottom .= "<tr><td align=right><b><small>Attachment:</small></b></td><td><a href=\"$bpath\" target=\"_self\"><small>".CGI::escapeHTML($imageName{$tag})."</small></a><br></td></tr>";
      $msgbottom .= "<tr><td align=right><b><small>File Size:</small></b></td><td><small>".CGI::escapeHTML($imageSize{$tag})." K</small></td></tr>";
      $msgbottom .= "</table>";
    }
  }
  print $msgbottom;

  print html_message_bottom($q);
}
 




######
# Recursive routine for dumping an entity.
######
sub dump_entity($$) {
  my ($entity, $name) = @_;
  my $IO;

  my($from, $cc, $to, $subject, $date);

  # Get any mail headers
  my $head = $entity->head;
  $from = $head->get('From') || '(Missing)';
  $subject = $head->get('Subject') || '(No subject)';
  $date = $head->get('Date') || '(Missing)';
  $cc = quote_from($q, $head->get('CC'));
  $to = quote_from($q, $head->get('To')) || '(Missing)';

  # Print mail headers nicely

  if ($date ne '(Missing)') {
    $from = quote_from($q, $from);
    $date = quote_date($q, $date);
    $subject = quote_subject($q, $subject);

    print html_message_headers($q, $from, $subject, $to, $cc, $date);
  }

  # Output the body:
  my @parts = $entity->parts;
  if (@parts) {                     # multipart...
    foreach my $i (0 .. $#parts) {       # dump each part...
      dump_entity($parts[$i], $name . ", part " . $i);
    }
  }
  else {                            # single part...  
    
    # Get MIME type, and display accordingly...
    my ($type, $subtype) = split('/', $entity->head->mime_type);
    my $body = $entity->bodyhandle;

    #if ( ($type =~ /^(text|message)$/) && (! ($entity->head->mime_attr('content-disposition') =~ /^inline$/i)) ) {     # text: display it...
    if ( (($type =~ /^(text|message)$/) && ($msgtypes{$subtype} != 1)) && ( (!($entity->head->mime_attr('content-disposition') =~ /^inline$/i)) || (scalar(%msgtypes) == 0) ) ) {
      # text & we haven't seen other instance of this subtype
      # (e.g., mail clients send the message first, then attachments)
      # : display it...
      if ($entity->head->mime_attr('content-disposition') =~ /^inline$/i) {
        $noInline = 0;
      }
      if ($IO = $body->open("r")) {
	my $txt;
	$txt .= $_ while (defined($_ = $IO->getline));
	$IO->close;

	if ($subtype eq 'html') { 
      # hey, HTML

	  # Should we display raw HTML?
	  if (force_escape_html() == 1) {
		$txt = CGI::escapeHTML($txt);
	  } elsif (force_escape_html() >= 2 ) {
		# clean according to safe_html_list()
		$txt = allow_safe_tags($txt);
	  } else {
	  	# Should we defang?
	  	$txt = defang($txt) if force_defang_html();
	  }

	  my $contentbase = $entity->head->get('Content-Base'); 

	  # Content base means that we've got an HTML page here, not
	  # just HTML. We need to mangle the links so that it thinks
	  # it's back home. Nasty. JavaScript may not work.
	  if ($contentbase) {
	    $contentbase =~ s|\"||g;
	    $txt = fix_links($txt, $contentbase);
	  }

          if (acme_anonymize_links()) {
            $txt = anonymize_links($q,$txt);
          }

	} else { 
      # oh, text but not HTML, quote everything, magic URLs, add BRs to each line...

	  # Should we defang? (really paranoid here, I know)
	  #$txt = defang($txt) if force_defang_html();

	  $txt = CGI::escapeHTML($txt);
	  $txt =~ s|^[A-Z]*&gt;.*|html_quoted($q, $&)|egm;
	  $txt =~ s|\n|<br>\n|g;	  
	  if (auto_add_hyperlinks()) {
	    $txt =~ s|(https?://[\[\]:a-z\-0-9/~._,\#=;?&%]+[a-z\-0-9/_~])|<a href="$1"><tt>$1</tt></a>|gim;
	    $txt =~ s|([^/>])(www\.[\[\]:a-z\-0-9/~._,\#=;?&]+\.[a-z\-0-9/_~]+)|$1<a href="http://$2"><tt>$2</tt></a>|gim;
	    $txt =~ s|<a href="(.+?)&gt"><tt>(.+?)&gt</tt></a>;|<a href="$1"><tt>$2</tt></a>&gt;|gim;
	    $txt =~ s|(ftp://[a-z\-0-9/~._,]+[a-z\-0-9/_~])|<a href="$1"><tt>$1</tt></a>|gim;
	  }

	  # Fancy up the signature
	  if ( $txt =~ /<br>\n--\s*<br>\n/s ) { 
	    $txt =~ s/<br>\n--\s*<br>\n(.*)$//s;
	    my $sig = $1; $sig =~ s:<br>::g;
	    $txt .= html_signature($q, $sig);
	  }
          if (acme_anonymize_links()) {
            $txt = anonymize_links($q,$txt);
          }
          # if $mb eq '1', we're in a frameset and don't have 
          # credentials for composing a message
          if ($mb ne '1') {
	    $txt =~ s|([a-z\-0-9._,]+\@[a-z.0-9]+[a-z0-9\-]*\.[a-z]{2,4})|"<tt>".quote_from_single_compose($q, $1, $1)."</tt>"|egim;
          }
	}

	# And finally print it
	$msgtops{$subtype} .= $txt;
	$msgtypes{$subtype} = 1;
      }
      else {       # d'oh!
	$msgtops{'plain'} .= "<h3>MIME Error</h3>MIME entity error - not found, sorry.";
      }
    }
    elsif (($type eq "image") && ($subtype =~ /^(gif|jpeg|png)$/) && ($entity->head->mime_attr('content-disposition') =~ /^inline$/i) && defined($entity->head->mime_attr('content-id')) ) { 
      # image, inline it...
      my $filename = 'cid-'.$attachNum.'-'.$origmailbox.'-'.$show.'.'.$subtype;
      ++$attachNum;
      my $itag = 'cid'.$entity->head->mime_attr('content-id');
      $inlineImages{$itag} = $subtype;
      $imageName{$itag} = $entity->head->recommended_filename || "no-name.$subtype";
      $imageName{$itag} =~ s/^.*\/([^\/]*)$/$1/;
      my $path = CGI::escapeHTML(web_temporary_mime_directory().$sessionid.'/'.acme_file_escape($filename));
      $imageURI{$itag} = $path;
      my $fpath = temporary_mime_directory().$sessionid.'/'.acme_file_escape($filename);
      my $bod = $entity->bodyhandle;
      my $data = '';
      if ($IO = $bod->open("r")) {
	my $x;
        while ($IO->read($x,2048)) { $data .= $x ; }
        $IO->close;
      }
      if (open(D,">$fpath")) { print D $data; close D; }
      my $size;
      my @finfo = stat($fpath);
      if ($? == 0) {
	$size = int($finfo[7] / 1024);
	$size =~ s/^0$/\&lt;1/;
      } else {
	$size = "???"; # something strange
      }
      $imageSize{$itag} = $size;
    }
    else {                                 
      # binary: just summarize it...
      my $fname = $entity->head->recommended_filename || 'no-name';
      $fname =~ s/^.*\/([^\/]*)$/$1/;
      my $filename = 'file-'.$attachNum.'-'.$fname;
      ++$attachNum;
      my $q2 = new CGI($q);
      $q2->delete_all();
      $q2->param('att',$filename);
      $q2->path_info('/view/'.$fname);
      my $md5 = new Digest::MD5;
      $md5->add($filename,$messageSecret);
      $q2->param('msgkey',$md5->hexdigest());
      if (!defined($q->cookie('session-id'))) { $q2->param('sessionid',$sessionid); }
      my $path = CGI::escapeHTML($q2->url(-full=>1, -query=>1, -path_info=>1));
      my $fpath = dope_directory().$sessionid.'-mime/'.acme_file_escape($filename);
      my $bod = $entity->bodyhandle;
      my $data = '';#$entity->Data;
      if ($IO = $bod->open("r")) {
	my $x;
        while ($IO->read($x,2048)) { $data .= $x ; }
        $IO->close;
      }
      if (open(D,">$fpath")) {
        my $mtype = $type."/".$subtype;
        $mtype =~ s/\0//g;
        print D $mtype . chr(0);
        print D $data;
        close D;
      }

      my $size = length($data);
      $size = int($size / 1024);
      $size =~ s/^0$/\&lt;1/;

      $msgbottom .= "<table bgcolor=\"#cccccc\" cellspacing=0 cellpadding=2 border=0%>";
      $msgbottom .= "<tr><td align=right><b><small>Attachment:</small></b></td><td><a href=\"$path\" target=\"_self\"><small>".CGI::escapeHTML($fname)."</small></a><br></td></tr>";
      $msgbottom .= "<tr><td align=right><b><small>MIME Type:</small></b></td><td><small>".CGI::escapeHTML($type."/".$subtype)."</small></td></tr>";
      $msgbottom .= "<tr><td align=right><b><small>Size:</small></b></td><td><small>${size}K</small></td></tr>";
      $msgbottom .= "</table>";
    }
  }
}

# return a safe filename, basically URI-escaping unexpected chars
sub acme_file_escape($) {
  my $srcname = shift;
  $srcname =~ s/([^a-z0-9\-.\_\,\(\)\;\'\-])/sprintf('\%%x',ord($1))/gei;
  return($srcname);
}

# routine to modify IMG tags with SRC attributes that point to inline images
sub acme_fix_cid_src($) {
  my $txt = shift;
  my($buffer,$tag);
  while ( $txt =~ s/^(.*?)\<img\s*src\s*=\s*\"cid\:([^\"]*?)\"(.*)$/$3/si ) {
    $buffer .= $1; $tag = $2;
    $tag = 'cid<'.$tag.'>';
    $shownImages{$tag} = 1;
    $tag = $imageURI{$tag};
    $buffer .= "<img src=\"$tag\" ";
  }
  $buffer .= $txt;
  return($buffer);
}

# routine to anonymize A HREF links
sub anonymize_links($$) {
  my ($q2,$txt) = @_;
  my $q = new CGI($q2);
  my($buffer,$tag,$attr);
  $q->delete_all();
  $q->path_info('');
  while ( $txt =~ s/^(.*?)\<a\b(.*?)\bhref\s*=\s*\"(https?:[^\"]*?)\"(.*)$/$4/si ) {
    $buffer .= $1; $attr = $2; $tag = $3;
    $q->param('aredir',"$tag");
    $buffer .= "<a $attr href=\"$tag\" ";
  }
  $buffer .= $txt;
  $txt = $buffer;
  $buffer = '';
  while ( $txt =~ s/^(.*?)\<a\b(.*?)\bhref\s*=\s*(https?:[^\s]*?")(\s.*)$/$4/si ) {
    $buffer .= $1; $attr = $2; $tag = $3;
    $q->param('aredir',"$tag");
    $buffer .= "<a $attr href=\"$tag\" ";
  }
  $buffer .= $txt;
  return($buffer);
}

#####
# Compose a message
#####

sub compose($$$$$$$$$$) {
  my($q, $cclient, $to, $cc, $bcc, $inreplyto, $subject, $attachment, $attachment_type, $body) = @_;

  # protect everything from getting trashed when it's passed in html
 
  encode_entities($to);
  encode_entities($cc);
  encode_entities($subject);
  encode_entities($body);

  # create the form  (Eventually needs to go to AcmemailHtmlConfig.pm or whatever)

  $q->path_info('/control/');
  print $q->start_multipart_form();
  print qq|
<table cellspacing="0" cellpadding="2" border="0" bgcolor="#000000"><tr><td>
<table cellspacing="0" cellpadding="3" border="0" bgcolor="#ffffff"><tr><td>
<table>
<tr>
 <td align="center" colspan="2"><small>Separate email addresses with a comma.</small></td>
</tr>
<tr>
 <td align="right" valign="center">
  <span class="mheading"><b>To:</b></span>
 </td>
 <td>
  <INPUT TYPE="text" NAME="to" SIZE="60" VALUE="$to">
 </td>
</tr>
<tr>
 <td align="right" valign="center">
  <span class="mheading"><b>Cc:</b></span>
 </td>
 <td>
  <INPUT TYPE="text" NAME="cc" SIZE="60" VALUE="$cc">
 </td>
</tr>
<tr>
 <td align="right" valign="center">
  <span class="mheading"><b>Bcc:</b></span>
 </td>
 <td>
  <INPUT TYPE="text" NAME="bcc" SIZE="60" VALUE="$bcc">
 </td>
</tr>
<tr>
 <td align="right" valign="center">
  <span class="mheading"><b>Attachment:</b></span>
 </td>
 <td>
  <INPUT TYPE="file" NAME="attachment" SIZE="46">
 </td>
</tr>
<tr>
 <td align="right" valign="center">
  <span class="mheading"><b>Subject:</b></span>
 </td>
 <td>
  <INPUT TYPE="text" NAME="subject" SIZE="60" VALUE="$subject">
 </td>
</tr>
<tr>
 <td colspan="2">
  <TEXTAREA NAME="body" COLS="70" ROWS="20">$body</TEXTAREA>
 </td>
</tr>
</table>
<INPUT TYPE="submit" NAME="send" VALUE="Send">
|;
if ( (!force_cookies()) && (!defined($q->cookie('cantakecookies'))) ) {
  if (!defined($q->cookie('session-id'))) {
    print $q->hidden('sessionid',$sessionid);
  }
  if ( (!defined($q->cookie('key'))) && encrypt_password() ) {
    print $q->hidden('key',$passwordPad);
  }
  if( (!defined($q->cookie('control'))) && ($q->path_info() eq '/control/') ) {
    print $q->hidden('control',$controlValue);
  }
}
  print $q->hidden('inreplyto', $inreplyto);
  print qq|
</form>
</td></tr></table>
</td></tr></table>
|;
}




######
# Send a message
######

sub mail($$$$$$$$$$$$$) {
  my($q, $cclient, $mailbox, $uid, 
     $username, $to, $cc, $bcc, $inreplyto, $subject, $attachment, $attachment_type, $body) = @_;

  if (not $to) {
    print qq|<b>Please enter a recipient for this message</b><br>|;
    compose($q, $cclient, $to, $cc, $bcc, $inreplyto, $subject, $attachment, $attachment_type, $body);
    return;
  } elsif (not $subject) {
    print qq|<b>Please enter a subject for this message</b><br>|;
    compose($q, $cclient, $to, $cc, $bcc, $inreplyto, $subject, $attachment, $attachment_type, $body);
    return;
  } elsif (not $body) {
    print qq|<b>Please enter some message text</b><br>|;
    compose($q, $cclient, $to, $cc, $bcc, $inreplyto, $subject, $attachment, $attachment_type, $body);
    return;
  }
    
  $body =~ s/\r\n/\n/g;   # Some browsers send CRLF - drop extra CRs

  my $top = MIME::Entity->build(To         => $to,
				Subject    => $subject,
				From       => $username . '@' . local_domain(),
				'Reply-To' => $username . '@' . local_domain(),
				Cc         => $cc,
				Bcc        => $bcc,
				'In-Reply-To:' => ( ($inreplyto ne '') ? $inreplyto : undef),
				'Content-Type' => 'text/plain; charset=us-ascii',
				Data       => [$body],
				Date       => Mail::Cclient::rfc822_date(),
				'X-Mailer' => 'acmemail <URL:http://www.astray.com/acmemail/>'
			       );

  if (defined $attachment && (length($attachment) > 0)) {
    $top->attach(Data     => $attachment,
		 Type     => $attachment_type,
		 Encoding => '-SUGGEST');
  }

  my $smtp = Net::SMTP->new(smtphost());

  if (defined $smtp) {
    $smtp->mail($username . '@' . local_domain());
    my @rcptto = split(/\s*,\s*/, $to);
    my @rcptcc = split(/\s*,\s*/, $cc);
    my @rcptbcc = split(/\s*,\s*/, $bcc);
    my @rcpt = (@rcptto, @rcptcc, @rcptbcc);
    $smtp->to(@rcpt);
    $smtp->data();
    $smtp->datasend($top->stringify);
    $smtp->dataend;
    $smtp->quit;
    
    print qq|<b>The mail was sent off!</b><br>|;
    list($q, $cclient, $mailbox, $offset);
  } else {
    print qq|<b>Error: the mail server could not be contacted. Mail not sent.</b>|;
    compose($q, $cclient, $to, $cc, $bcc, $inreplyto, $subject, $attachment, $attachment_type, $body);
  }
}








######
# Delete a message
######

sub trash($$@) {
  my($q, $cclient, @todelete) = @_;

  foreach (@todelete) {
    # Mark each message deleted
    $cclient->setflag($_, '\\Deleted', 'uid');
  }

  # And actually do the deleting

  $cclient->expunge;
#  if ($cclient->expunge) {
    if (@todelete == 1) {
      print "<b>Message deleted!</b>";
    } else {
      print "<b>Messages deleted!</b>";
    }
#  } else {
#    print "<b>Unable to delete message!</b>";
#  }

  print "<p>";
  # And then list() is called
}





####
# Forward a message
####

sub forward($$$) {
  
  my ($q, $cclient, $uid) = @_;
  my ($body, $subject); 

  # rather than doing a 'proper' forward, we simply extract the
  # correct details from the old email so they can be resent

  # we extract the details directly as if they'd been passed to us
  if ($uid ne '' ) {
    $subject = $cclient->fetchheader($uid, ['Subject'],'uid'); 
    $subject =~ s|^Subject[\s]*: +|(fwd) |;
    $body = "\n".$cclient->fetchbody($uid, 1, 'uid');

    $body = quote_body($cclient, $uid, "");
  }
  # eventually this method will be completely superceeded with
  # the forwarding of the whole last message as a mime attachement

  # now simply compose as we could have done anyhow
  compose($q, $cclient, $to, $cc, $bcc, $inreplyto, $subject, $attachment, $attachment_type, $body);
}



####
# Reply to a message
####
sub reply($$$$) {

  my($q, $cclient, $uid, $replytoall) = @_;
  my($to, $body, $subject, $cc, $inreplyto);  

  # we extract the details directly as if they'd been passed to us
  if ($uid ne '') {

    # work out who the mail was from
    $_ = $cclient->fetchheader($uid, ['Resent-Reply-To'], 'uid'); 
    $_ = $cclient->fetchheader($uid, ['Resent-From'],'uid')     if /^\s*/;
    $_ = $cclient->fetchheader($uid, ['Reply-To'],'uid')        if /^\s*/;
    $_ = $cclient->fetchheader($uid, ['From'],'uid')            if /^\s*/;
    s/^[^:]*: +//;                   # remove the fieldname    
    $to = $_;
    $to =~ s|/r| |g;    # ugh, windoze
    $to =~ s/\s+/ /g;   # 'unfold' as RFC822 likes to call it
    
    # add make subject line
    $_ = $cclient->fetchheader($uid, ['Resent-Subject'],'uid');
    $_ = $cclient->fetchheader($uid, ['Subject'],'uid')         if /^\s*/;

    s|^Subject[\s]*: +|Re: |i;   # remove the fieldname, add a 'Re'
    s|^(re\s*:\s*){2,}|Re: |i;   # remove multiple 'Re's
    $subject = $_;

    # get message id
    $_ = $cclient->fetchheader($uid, ['Message-Id'],'uid');
    s|^Message-Id[\s]*: +||i;	# remove the fieldname
    s|\s*$||i;
    $inreplyto = $_;

    # add everyone if we are replying to all
    if ($replytoall) {
      $_ = $cclient->fetchheader($uid, ['Resent-Cc'],'uid');
      $_ = $cclient->fetchheader($uid, ['Cc'],'uid')            if /^\s*/;     
      s/^[^:]*: +//;                   # remove the fieldname   
      $cc = $_;
      
      $_ = $cclient->fetchheader($uid, ['Resent-To'],'uid');
      $_ = $cclient->fetchheader($uid, ['To'],'uid')            if /^\s*/;     
      s/^[^:]*: +//;                   # remove the fieldname 
      $cc .= ", $_";
      
      # remove all instances of ourselves
      # currently this is limited to our replying address, but wait for address books
      # this doesn't cover commas in people's name yet
      my $myself = $username . '@' . local_domain();
      $cc =~ s/[^,]*[<\s]+$myself[>\s]+[^,]*//gi;
      
      $cc =~ s|/r||g;      # ugh, windoze
      $cc =~ s/\s+/ /gs;   # 'unfold' as RFC822 likes to call it
      
      # clear up excess commas
      $cc =~ s/[,]\s*[,]//g;   # double commas away
      $cc =~ s/\s*[,]\s*$//g;  # commas at the end away
      $cc =~ s/^\s*[,]\s*//g;  # commas at the start away
    }
    
    # deal with the body 
    $body = quote_body($cclient, $uid, $to);
  }  

  # now simply compose as we could have done anyhow (no replying of attachments as of yet - not efficent)
  compose($q, $cclient, $to, $cc, undef, $inreplyto, $subject, undef, undef, $body);
}








######
# Mailboxes (return a list of possible mailboxes)
######

sub mailboxes($) {
  my $cclient = shift;

  my %seen;
  my $inbox = inbox_directory();

  set_callback "list" => sub {
    shift; # drop mailstream argument
    shift; # drop separator character
    my $name = shift;
    $name =~ s/^.*}//;
    return if substr($name, 0, 1) eq "." || $seen{$name}++;
  };
 
  $cclient->list('{' . mailhost() . '/' . mailhost_type() . '}', $inbox . '/%');

  my @mailboxes = map { 
    if ($_ eq ($inbox . '/')) {
      'INBOX';
    } else {
      $_ =~ s|^$inbox/||;
      $_;
    }
  } sort { lc $a cmp lc $b } keys %seen;
  return 'INBOX' unless @mailboxes;
  return @mailboxes;
}



######
# Loosedir (delete a directory)
######

sub loosedir($) {
  my $dir = shift;
  foreach my $file (<$dir/*>) {
    unlink $file 
  }
  rmdir $dir;
}






######
# Mangle HTML links
######

# Change HTML links in $content to point back to under $url
# Copes with links, graphics, forms, frames, stylesheets, local links (#)...
# Note: this is nasty. Don't try to understand.

sub fix_links($$) {

  my ($content, $url) = @_;
  my ($tag, $parm, $parms, $uri, $origuri, $want);
  my $local = "";
  
  $content =~ s|[\n\r]| |g; # Temporary hack
  $content =~ s|\<base.*?\>||i; # No, no, no! No BASE tags!
  
  my $new_content = $content;
  
  while ($content =~ m|<|) {
    
    $content =~ s|^.*?\<(.*?)([> ])||so; # Get rid of anything until first tag
    $tag = lc $1;

    next if ($2 eq ">"); # Get rid of empty tags

    $content =~ s|^(.*?)>||so; # Until end of tag
    $parms = $1;
    $parm = "";
    $want = "";
    
    if ($tag eq "a") { $want = "href"; }
    elsif ($tag eq "frame") { $want = "src"; }
    elsif ($tag eq "img") { $want = "src"; }
    elsif ($tag eq "form") { $want = "action"; }
    elsif ($tag eq "body") { $want = "background"; }
    elsif ($tag eq "link") { $want = "href"; }
    elsif ($tag eq "input") { $want = "src"; }
    
    next if ($want eq "");
    
    if ($parms=~m|^.*?($want *= *\"?(.*?)[\"\s>]).*$|si) {
      $parms = $1;
      $parm = $2;
    }

    next if ($parm eq "");
    
    $local = "";
    $local = $1 if ($parm =~ s|(\#.+)$||); # Get rid of local references

    $uri = (URI::WithBase->new($parm, $url))->abs;

    next if ($uri->scheme ne "http");
    
    $origuri = $uri->as_string;

    $parm = $uri->as_string;
    $parm .= "index.html" if (basename($parm eq "")); # Hack!
    $parm .= $local; # Cope with local references
    
    $parms = quotemeta($parms); # NEW!
    
    $new_content =~ s|$parms|$want=\"$parm\"|g; # And insert it...
  }
  return $new_content;
}







# Quote a mail header
sub quote_header($) {
  my $header = shift;

  if ($header) {

    if ($header =~ s|=\?iso-8859-1\?Q\?||) {
      $header =~ s|\?=||g;
      $header =~ s|=(..)|chr(hex($1))|ge;
    }

    $header =~ s|&|&amp;|mg;
    $header =~ s|\"|&quot;|mg;
    $header =~ s|<|&lt;|mg;
    $header =~ s|>|&gt;|mg;
    chomp $header;
  }
  $header =~ s|\s+$||g;
  return $header;
}


# Quote a subject
sub quote_subject($$) {
  my($q, $subject) = @_;
  $subject = quote_header($subject);

  $subject = "(No subject)" if ($subject eq ""); # Duh, silly people (Mark!)

  return $subject;
}
    

# Match comma-delimited email addresses and pass 
# them to quote_from_single
sub quote_from($$) {
  my($q, $from) = @_;

  if ($from) {
    if (($from =~ m|\@.+\@|) && ($from =~ m|, |)) {
      my @froms = split ',', $from;
      @froms = map quote_from_single($q, $_), @froms;
      return join ', ', @froms;
    } else {
      return quote_from_single($q, $from);
    }
  } else {
    return '<br>';
  }
}


# Quote a from/to/cc (adds compose-to's in most cases)
# Matches: Leon Brocard <leon@astray.com>
#          leon@astray.com (Leon Brocard)
#          leon@astray.com
sub quote_from_single($$) {
  my($q, $from) = @_;
  if (defined($show) && ($mb eq '1')) {
    # We're in a cookie-free frameset and don't have credentials
    # to send a message
    return(CGI::escapeHTML($from));
  }
  $from = quote_header($from);

  $from =~ s|^\s+||;
  $from =~ s|\s+$||;
  $from =~ s|\"||g;
  $from =~ s|\n||g;

  if ($from =~ s|^&quot;(.+?)&quot; &lt;(.+)&gt;$||i) {
    $from = quote_from_single_compose($q, $2, $1);
  } elsif ($from =~ s|^'?(.+?)'? &lt;(.+)&gt;$||i) {
    $from = quote_from_single_compose($q, $2, $1);
  } elsif ($from =~ s|^(.+) \((.+)\)$||i) {
    $from = quote_from_single_compose($q, $1, $2);
  }  elsif ($from =~ m|^(.+\@.+)$|) {
    $from = quote_from_single_compose($q, $1, $1);
  }

  $from =~ s|^\"||;
  $from =~ s|\s$||;
  $from =~ s|\"$||;

  $from = "<br>" if ($from eq "");

  return $from;
}


# Create a 'mailto'-like URL, but with acmemail's
# compose instead
sub quote_from_single_compose($$$) {
  my($q, $email, $name) = @_;

  my $newemail = "$email ($name)";
  $name .= " &lt;$email&gt;" if (!hide_email_address() && $email ne $name);
  $q->path_info('/control/');
  $q->param('compose', 1);
  $q->param('to', $newemail);
  my $url = $q->self_url;
  $q->delete('compose');
  $q->delete('to');
  return qq|<a href="$url">$name</a>|;
}
			       


# Quotes a date, putting it in absolute format if it's more than
# a week old, in Weekday Date Time if it's less than a week and
# just the time if it's dated today.
sub quote_date($$) {
  my($q, $date) = @_;
  $date = quote_header($date);
  $date = str2time($date);

  if (length($date) < 10) {
	return "";
  }

  # Get the hour and minute of current time
  my $now = time;
  my $hour= time2str("%H", $now);
  my $minute = time2str("%M", $now);

  if ($date > ($now - (($minute * 60) + ($hour * 60 * 60)))) {
    $date = time2str("%R", $date);
  } elsif ($date > ($now - (60*60*24*7))) {
    $date = time2str("%a %R", $date);
  } else {
    $date = time2str("%a %e/%m/%y %R", $date);
  }
  return $date;
}



####
# Quote up an old body
####
 
sub quote_body($$$) {
  my ($cclient, $uid, $from) = @_;
  
  # keep it simple for now.  What this actually should do is :
  
  # 1. quote up the plain text body if there
  # 2. otherwise quote the multipart/alternative
  # 3. otherwise strip the html of *everything* and use that quoted up
  
  # so far we only do number one.
  
  # who wrote the mail? - doens't display multiple froms very nice
  $_ = $cclient->fetchheader($uid, ['Resent-Sender'],'uid');
  $_ = $cclient->fetchheader($uid, ['Sender'],'uid')          if /^\s*/;
  $_ = "Default: $from"                                    if /^\s*/;
  s/^[^:]*: +//;  # remove fieldname
  my $somebody = $_;
  
  # when did they write it?
  $_ = $cclient->fetchheader($uid, ['Date'],'uid');
  s/^[^:]*: +//;             # remove fieldname
  my $somewhen = $_;
  
  my $preamble = "At ${somewhen}, $somebody wrote:";
  $preamble =~ s|/r| |g;     # ugh, windoze
  $preamble =~ s/\s+/ /gs;   # 'unfold' as RFC822 likes to call it
  
  my $body = $cclient->fetchbody($uid, 1, 'uid');           # get the body
  $body =~ s|^|>|s;                                     # quote the message with '>'
  $body =~ s|\n|\n>|sg;                                 # quote the message with '>'
  $body = "$preamble \n\n$body";                        # add the preamble
  
  $body;           # make sure we return the right thing.
}



# Try to remove malicious HTML
# (code originally by John D. Hardin)
sub defang {
  my $txt = shift;
  my($saved, $char);

  $_ = $txt;

  # First things first, unobfuscate, e.g. "&#116;" -> "t"
  if (/["\047][^"\047\s]*&#x?[1-9][0-9a-f]/i) {
      while (/["\047][^"\047\s]*&#((4[6-9]|5[0-8]|6[4-9]|[78][0-9]|9[07-9]|1[0-1][0-9]|12[0-2]))/) {
	     $char = chr($1);
	     s/&#$1;?/$char/g;
	   }
      while (/["\047][^"\047\s]*&#(x(2[ef]|3[0-9a]|4[0-9a-f]|5[0-9a]|6[1-9a-f]|7[0-9a]))/i) {
	     $char = chr(hex("0$1"));
	     s/&#$1;?/$char/gi;
	   }
    }

  if (/["\047][^"\047\s]*%[2-7][0-9a-f]/i) {
    while (/["\047][^"\047\s]*%((2[ef]|3[0-9a]|4[0-9a-f]|5[0-9a]|6[1-9a-f]|7[0-9a]))/i) {
      $char = chr(hex("0x$1"));
      s/%$1/$char/gi;
    }
  }

  # Now we can look for hostile tags
  if (/</) {
    s/<(META|APP|SCRIPT|OBJECT|EMBED|FRAME|IFRAME|STYLE)(\s|>)/<DEFANGED_$1$2/gi;

    s/STYLE\s*=/DEFANGED_STYLE=/gi;
    s/On(Abort|Blur|Change|Click|DblClick|DragDrop|Error|Focus|KeyDown|KeyPress|KeyUp|Load|MouseDown|MouseMove|MouseOut|MouseOver|MouseUp|Move|Reset|Resize|Select|Submit|Unload)/DEFANGED_On$1/gi;
  }

  s/(["\047])([a-z]+script|mocha|position|visibility):/${1}DEFANGED_$2:/gi;
  s/(["\047])&{/${1}DEFANGED_&{/g;

  # mailto/news/about/chrome/wysiwyg URLs in IMG SRC tags
  s/\<img\b(.*?\bsrc\s*?\=\s*)(mailto|news|about|chrome|wysiwyg)(\:.*?\>)/<IMG ${1}defanged_${2}${3}/gi;

  return $_;
}

# are we missing the control value?
sub control_cookie_problem() {
  my $md5 = new Digest::MD5;
  $md5->add($controlValue);
  if ( $md5->hexdigest() ne $controlHash ) { return(1); }
  return(0);
}

# referer_problem: decide if there's a problem with the Referer header
sub referer_problem ($$$$) {
	my ($actual,$isDangerous,$test,$refer) = @_;
	my $rc = 1;
	# test == 0 means do not check
	if ($test == 0) { return 0; }
	# test == 1 means only test if dangerous
	if ( ($test == 1) && ($isDangerous == 0) ) { return 0; }
	# this means we need to test!
	# lowercase the host/protocol of $actual because those
	# parts of the URL are not case-sensitive (but the rest is)
	$actual =~ m/^([^\/]*?\/\/[^\/]*?\/)(.*)$/;
	$actual = lc($1).$2;
	if ( substr($actual,0,length($refer)) eq $refer ) { $rc = 0; }
	return $rc;
}

# msgkey_problem: did the request present a valid msgkey?
sub msgkey_problem() {
  my $md5 = new Digest::MD5;
  $md5->add($show,$messageSecret,$origmailbox);
  if ( $md5->hexdigest() ne $msgkey ) { return(1); }
  return(0);
}

# allow_safe_tags: escape a body of HTML except for known good tags
sub allow_safe_tags($) {
	my $txt = shift;
	my $newtext = '';
	my $numStrippedTags = 0;
	my %safeList =  safe_html_list();
	my ($tag,$post,$which,$attrs,$good,$goodattr,$onlist,$attr,$id);
	while ( $txt =~ s/^(.*?)\<([^\>]*?)\>(.*)$/$3/s ) {
		$newtext .= CGI::escapeHTML($1);
		$tag = $2; $post = $3;
		if ( $tag =~ /^(\/?[a-z0-9]{1,})\b\s*(.*)\s*$/i ) {
			$which = $1; $attrs = $2; $onlist = 0;
			TTEST: foreach $good (keys %safeList) {
				if ( $which =~ /^\/?${good}$/i ) { $onlist = 1; $id = $good; last TTEST; }
			}
			if ( $onlist == 1 ) {
				$newtext .= '<' . $which;
				if ( $attrs ne '' ) {
				  # are attrs OK? 
				  foreach $goodattr ( @{ $safeList{$id} } ) {
				    while ( $attrs =~ s/^(.*?)(${goodattr})(.*)$/$1 $3/s ) {
					$newtext .= " $2";
				    }
				  }
				}
				$newtext .= '>';
			} else {
			   if ( force_escape_html() == 2 ) { 
				$newtext .= CGI::escapeHTML('<'.$tag.'>');
			   }
			}
		} else {
		   if ( force_escape_html() == 2 ) { 
			$newtext .= CGI::escapeHTML('<'.$tag.'>');
		   } else {
			++$numStrippedTags;
		   }
		}
	}
	$newtext .= CGI::escapeHTML($txt);
	# un-escape SGML entities
	$newtext =~ s/\&amp\;(\#?\w*)\;/\&$1\;/sg;
	if ( (force_escape_html() != 2)  && ($numStrippedTags > 0)) { 
		$newtext .= "<hr><b>acmemail: $numStrippedTags unknown/unsafe HTML tags were removed</b><br>\n";
	}
	return($newtext);
}

# There's nothing else here, go away!
