#!/usr/bin/perl -w

#
# The program is a converter between the XML format used for the
# documentation.xml file and HTML. That is, an HTML version of the
# documentation can be generated by:
#   ./xml.pl > documentation.html
#
# A plain text version of the documentation can be generated using
# something like:
#   lynx -dump documentation.html > documentation.txt
#
# Written by Leon Brocard / acme@astray.com

use strict;
use XML::XPathScript;

open(IN, "documentation.xml");
my $xml = join '', <IN>;
close(IN);

my $stylesheet = q|
<html>
<head>
<title><%= findvalue("/document/title/text()") %></title>
</head>
<body>

<h1><%= findvalue("/document/title/text()") %></h1>
<%= findvalue("/document/author/text()") %><br>
<i><%= findvalue("/document/date/text()") %></i>

<p><h2>Index</h2><p>

<ul>
<%
my $count = 1;
for my $sect (findnodes("/document/section")) {
  print "<li>$count: " . $sect->findvalue("\@name"), "<br>\n";
  my $subcount = 1;
  for my $subsect ($sect->findnodes("subsection")) {
    print "<li>$count.$subcount: " . $subsect->findvalue("\@name"), "<br>\n";
    $subcount++;
  }
  $count++;
}
%>
</ul>

<p>

<%
$t->{'list'}{pre} = '<ul>';
$t->{'list'}{post} = '</ul>';

$t->{'item'}{pre} = '<li>';
$t->{'item'}{post} = '</li>';

$t->{'coderef'}{pre} = '<tt>';
$t->{'coderef'}{post} = '</tt>';

$t->{'link'}{pre} = '<a href="{@href}">{@description}</a>';

$t->{'section'}{pre} = '<p><h2>{@name}</h2><p>';

$t->{'subsection'}{pre} = '<p><h3>{@name}</h3><p>';

$t->{'code'}{pre} = '<table width="100%"><tr><td bgcolor="#DDDDDD"><pre>';
$t->{'code'}{post} = '</pre></td></tr></table>';
%>

<%
# display all the sections
foreach my $section (findnodes("/document/section")) {
	print apply_templates($section);
}
%>

</body>
</html>
|;

my $xps = XML::XPathScript->new(xml => $xml, stylesheet => $stylesheet);
$xps->process;
